#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['spylib', 'spylib.oauth', 'spylib.utils']

package_data = \
{'': ['*']}

setup(name='spylib',
      version='0.2.2',
      description="A library to facilitate interfacing with Shopify's API",
      author='Anthony Hillairet',
      author_email='ant@satel.ca',
      url='https://github.com/SatelCreative/satel-spylib',
      packages=packages,
      package_data=package_data,
     )
